import unittest
from urlchecker_bot import WebBot
import mock
from test_urlchecker_dns import MockAnswer
import urlchecker_dns


def checkNetworkChange(bot, result_list):
    assert bot._db.NetworkChange.select().count() == len(result_list)
    select_list = (
        bot._db.NetworkChange.select()
        .order_by(bot._db.NetworkChange.ip.asc())
        .order_by(bot._db.NetworkChange.port.asc())
    )
    assert [(x.ip, x.port) for x in select_list] == result_list


def checkHttpCodeChange(bot, result_list):
    assert bot._db.HttpCodeChange.select().count() == len(result_list)
    select_list = (
        bot._db.HttpCodeChange.select()
        .order_by(bot._db.HttpCodeChange.ip.asc())
        .order_by(bot._db.HttpCodeChange.url.asc())
    )
    assert [(x.ip, x.url) for x in select_list] == result_list


def checkDnsChange(bot, result_list):
    assert bot._db.DnsChange.select().count() == len(result_list)
    select_list = (
        bot._db.DnsChange.select()
        .order_by(bot._db.DnsChange.resolver_ip.asc())
        .order_by(bot._db.DnsChange.domain.asc())
    )
    assert [(x.resolver_ip, x.domain) for x in select_list] == result_list


class UrlCheckerBotTestCase(unittest.TestCase):
    def test_emptyConfiguration(self):
        resolver_ip = "192.168.0.254"

        resolver = urlchecker_dns.dns.resolver.Resolver(configure=False)
        resolver.nameservers.append(resolver_ip)

        with mock.patch(
            "urlchecker_configuration.get_default_resolver"
        ) as mock_get_default_resolver, mock.patch(
            "urlchecker_dns.dns.resolver.Resolver.query"
        ) as mock_query:
            mock_get_default_resolver.return_value = resolver

            mock_query.return_value = [MockAnswer("1.2.3.4")]
            bot = WebBot(mapping={"SQLITE": ":memory:"})
            bot.initDB()
            bot.iterateLoop()

            assert mock_query.call_count == 1

        assert bot._db.Status.select().count() == 1
        assert bot._db.Status.get().text == "loop"

        checkNetworkChange(bot, [(resolver_ip, 53)])

        checkDnsChange(bot, [(resolver_ip, "example.org")])

        checkHttpCodeChange(bot, [])

    def test_oneNameserverOneDomainOneIp(self):
        resolver_ip = "127.0.0.1"

        resolver = urlchecker_dns.dns.resolver.Resolver(configure=False)
        resolver.nameservers.append(resolver_ip)

        with mock.patch(
            "urlchecker_configuration.get_default_resolver"
        ) as mock_get_default_resolver, mock.patch(
            "urlchecker_dns.dns.resolver.Resolver.query"
        ) as mock_query, mock.patch(
            "urlchecker_network.socket.socket"
        ) as mock_socket, mock.patch(
            "urlchecker_http.request"
        ) as mock_request:

            mock_get_default_resolver.return_value = resolver
            mock_query.return_value = [MockAnswer("1.2.3.4")]

            bot = WebBot(
                mapping={"SQLITE": ":memory:", "DOMAIN": "example.org"}
            )
            bot.initDB()

            bot.iterateLoop()

            assert mock_query.call_count == 2
            assert mock_socket.call_count == 2
            assert mock_request.call_count == 2

        checkNetworkChange(
            bot, [(resolver_ip, 53), ("1.2.3.4", 80), ("1.2.3.4", 443)]
        )

        checkDnsChange(bot, [(resolver_ip, "example.org")])

        checkHttpCodeChange(
            bot,
            [
                ("1.2.3.4", "http://example.org"),
                ("1.2.3.4", "https://example.org"),
            ],
        )

    def test_twoSimilarNameserverOneDomainOneIp(self):
        resolver_ip = "127.0.0.1"
        resolver_ip_2 = "127.0.0.2"

        bot = WebBot(
            mapping={
                "SQLITE": ":memory:",
                "DOMAIN": "example.org",
                "NAMESERVER": "\n".join([resolver_ip, resolver_ip_2]),
            }
        )
        bot.initDB()

        with mock.patch(
            "urlchecker_dns.dns.resolver.Resolver.query"
        ) as mock_query, mock.patch(
            "urlchecker_network.socket.socket"
        ) as mock_socket, mock.patch(
            "urlchecker_http.request"
        ) as mock_request:

            mock_query.return_value = [MockAnswer("1.2.3.4")]

            bot.iterateLoop()

            assert mock_query.call_count == 4
            assert mock_socket.call_count == 2
            assert mock_request.call_count == 2

        checkNetworkChange(
            bot,
            [
                (resolver_ip, 53),
                (resolver_ip_2, 53),
                ("1.2.3.4", 80),
                ("1.2.3.4", 443),
            ],
        )

        checkDnsChange(
            bot, [(resolver_ip, "example.org"), (resolver_ip_2, "example.org")]
        )

        checkHttpCodeChange(
            bot,
            [
                ("1.2.3.4", "http://example.org"),
                ("1.2.3.4", "https://example.org"),
            ],
        )

    def test_oneNameserverTwoDomainOneIp(self):
        resolver_ip = "127.0.0.1"
        domain_1 = "example.org"
        domain_2 = "foo.example.org"

        bot = WebBot(
            mapping={
                "SQLITE": ":memory:",
                "DOMAIN": "\n".join([domain_1, domain_2]),
                "NAMESERVER": resolver_ip,
            }
        )
        bot.initDB()

        with mock.patch(
            "urlchecker_dns.dns.resolver.Resolver.query"
        ) as mock_query, mock.patch(
            "urlchecker_network.socket.socket"
        ) as mock_socket, mock.patch(
            "urlchecker_http.request"
        ) as mock_request:

            mock_query.return_value = [MockAnswer("1.2.3.4")]

            bot.iterateLoop()

            assert mock_query.call_count == 3
            assert mock_socket.call_count == 2
            assert mock_request.call_count == 4

        checkNetworkChange(
            bot, [(resolver_ip, 53), ("1.2.3.4", 80), ("1.2.3.4", 443)]
        )

        checkDnsChange(bot, [(resolver_ip, domain_1), (resolver_ip, domain_2)])

        checkHttpCodeChange(
            bot,
            [
                ("1.2.3.4", "http://%s" % domain_1),
                ("1.2.3.4", "http://%s" % domain_2),
                ("1.2.3.4", "https://%s" % domain_1),
                ("1.2.3.4", "https://%s" % domain_2),
            ],
        )

    def test_oneNameserverOneDomainTwoIp(self):
        resolver_ip = "127.0.0.1"
        domain = "example.org"

        bot = WebBot(
            mapping={
                "SQLITE": ":memory:",
                "DOMAIN": domain,
                "NAMESERVER": resolver_ip,
            }
        )
        bot.initDB()

        with mock.patch(
            "urlchecker_dns.dns.resolver.Resolver.query"
        ) as mock_query, mock.patch(
            "urlchecker_network.socket.socket"
        ) as mock_socket, mock.patch(
            "urlchecker_http.request"
        ) as mock_request:

            mock_query.return_value = [
                MockAnswer("1.2.3.4"),
                MockAnswer("1.2.3.5"),
            ]

            bot.iterateLoop()

            assert mock_query.call_count == 2
            assert mock_socket.call_count == 4
            assert mock_request.call_count == 4

        checkNetworkChange(
            bot,
            [
                (resolver_ip, 53),
                ("1.2.3.4", 80),
                ("1.2.3.5", 80),
                ("1.2.3.4", 443),
                ("1.2.3.5", 443),
            ],
        )

        checkDnsChange(bot, [(resolver_ip, domain)])

        checkHttpCodeChange(
            bot,
            [
                ("1.2.3.4", "http://%s" % domain),
                ("1.2.3.5", "http://%s" % domain),
                ("1.2.3.4", "https://%s" % domain),
                ("1.2.3.5", "https://%s" % domain),
            ],
        )

    def test_oneNameserverOneSubDomainOneIp(self):
        resolver_ip = "127.0.0.1"
        domain = "example.org"
        sub_domain = "foo.%s" % domain

        bot = WebBot(
            mapping={
                "SQLITE": ":memory:",
                "DOMAIN": sub_domain,
                "NAMESERVER": resolver_ip,
            }
        )
        bot.initDB()

        with mock.patch(
            "urlchecker_dns.dns.resolver.Resolver.query"
        ) as mock_query, mock.patch(
            "urlchecker_network.socket.socket"
        ) as mock_socket, mock.patch(
            "urlchecker_http.request"
        ) as mock_request:

            mock_query.return_value = [MockAnswer("1.2.3.4")]

            bot.iterateLoop()

            assert mock_query.call_count == 3
            assert mock_socket.call_count == 2
            assert mock_request.call_count == 4

        checkNetworkChange(
            bot, [(resolver_ip, 53), ("1.2.3.4", 80), ("1.2.3.4", 443)]
        )

        checkDnsChange(bot, [(resolver_ip, domain), (resolver_ip, sub_domain)])

        checkHttpCodeChange(
            bot,
            [
                ("1.2.3.4", "http://%s" % domain),
                ("1.2.3.4", "http://%s" % sub_domain),
                ("1.2.3.4", "https://%s" % domain),
                ("1.2.3.4", "https://%s" % sub_domain),
            ],
        )

    def test_oneNameserverOneUrlOneIp(self):
        resolver_ip = "127.0.0.1"
        domain = "example.org"

        bot = WebBot(
            mapping={
                "SQLITE": ":memory:",
                "URL": "https://example.org/foo",
                "NAMESERVER": resolver_ip,
            }
        )
        bot.initDB()

        with mock.patch(
            "urlchecker_dns.dns.resolver.Resolver.query"
        ) as mock_query, mock.patch(
            "urlchecker_network.socket.socket"
        ) as mock_socket, mock.patch(
            "urlchecker_http.request"
        ) as mock_request:

            mock_query.return_value = [MockAnswer("1.2.3.4")]

            bot.iterateLoop()
            assert mock_query.call_count == 2
            assert mock_socket.call_count == 2
            assert mock_request.call_count == 3

        checkNetworkChange(
            bot, [(resolver_ip, 53), ("1.2.3.4", 80), ("1.2.3.4", 443)]
        )

        checkDnsChange(bot, [(resolver_ip, domain)])

        checkHttpCodeChange(
            bot,
            [
                ("1.2.3.4", "http://%s" % domain),
                ("1.2.3.4", "https://%s" % domain),
                ("1.2.3.4", "https://example.org/foo"),
            ],
        )

    def test_oneNameserverOneUrlNoIp(self):
        resolver_ip = "127.0.0.1"
        domain = "example.org"

        bot = WebBot(
            mapping={
                "SQLITE": ":memory:",
                "URL": "https://example2.org/foo",
                "NAMESERVER": resolver_ip,
            }
        )
        bot.initDB()

        with mock.patch(
            "urlchecker_dns.dns.resolver.Resolver.query"
        ) as mock_query, mock.patch(
            "urlchecker_network.socket.socket"
        ) as mock_socket, mock.patch(
            "urlchecker_http.request"
        ) as mock_request:

            mock_query.side_effect = [[MockAnswer("1.2.3.4")], []]

            bot.iterateLoop()
            assert mock_query.call_count == 2
            assert mock_socket.call_count == 0
            assert mock_request.call_count == 0

        checkNetworkChange(bot, [(resolver_ip, 53)])

        checkDnsChange(
            bot, [(resolver_ip, domain), (resolver_ip, "example2.org")]
        )

        checkHttpCodeChange(bot, [])


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(UrlCheckerBotTestCase))
    return suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")
