import socket
import errno
from peewee import fn


TIMEOUT = 2


def reportNetwork(db, ip=None, transport=None, port=None):
    # http://charlesleifer.com/blog/techniques-for-querying-lists-of-objects-and-determining-the-top-related-item/
    NetworkChangeAlias = db.NetworkChange.alias()

    subquery = (
        NetworkChangeAlias.select(
            NetworkChangeAlias.ip,
            NetworkChangeAlias.transport,
            NetworkChangeAlias.port,
            fn.MAX(NetworkChangeAlias.status_id).alias("max_status_id"),
        )
        .group_by(
            NetworkChangeAlias.ip,
            NetworkChangeAlias.transport,
            NetworkChangeAlias.port,
        )
        .alias("network_change_max_subquery")
    )

    StatusAlias = db.Status.alias()

    query = (
        db.NetworkChange.select(db.NetworkChange, StatusAlias)
        .join(StatusAlias)
        .switch(db.NetworkChange)
        .join(
            subquery,
            on=(
                (db.NetworkChange.status_id == subquery.c.max_status_id)
                & (db.NetworkChange.ip == subquery.c.ip)
                & (db.NetworkChange.transport == subquery.c.transport)
                & (db.NetworkChange.port == subquery.c.port)
            ),
        )
        .order_by(
            db.NetworkChange.ip.asc(),
            db.NetworkChange.port.asc(),
            db.NetworkChange.transport.asc(),
        )
    )

    if transport is not None:
        if type(transport) == list:
            query = query.where(db.NetworkChange.transport << transport)
        else:
            query = query.where(db.NetworkChange.transport == transport)
    if port is not None:
        if type(port) == list:
            query = query.where(db.NetworkChange.port << port)
        else:
            query = query.where(db.NetworkChange.port == port)
    if ip is not None:
        if type(ip) == list:
            query = query.where(db.NetworkChange.ip << ip)
        else:
            query = query.where(db.NetworkChange.ip == ip)
    return query


def logNetwork(db, ip, transport, port, state, status_id):

    with db._db.atomic():
        try:
            # Check previous parameter value
            previous_entry = reportNetwork(
                db, ip=ip, transport=transport, port=port
            ).get()
        except db.NetworkChange.DoesNotExist:
            previous_entry = None

        if (previous_entry is None) or (previous_entry.state != state):
            previous_entry = db.NetworkChange.create(
                status=status_id,
                ip=ip,
                transport=transport,
                port=port,
                state=state,
            )
        return previous_entry.status_id


def isTcpPortOpen(db, ip, port, status_id, timeout=TIMEOUT):
    is_open = False
    sock = socket.socket()
    sock.settimeout(timeout)
    try:
        sock.connect((ip, port))
        state = "open"
        is_open = True
    except ConnectionRefusedError:
        state = "closed"
    except socket.timeout:
        state = "filtered"
    except OSError as e:
        if e.errno == errno.EHOSTUNREACH:
            # OSError: [Errno 113] No route to host
            state = "filtered"
        elif e.errno == errno.ENETUNREACH:
            # OSError: [Errno 101] Network is unreachable
            state = "unreachable"
        else:
            sock.close()
            raise
    except:
        sock.close()
        raise

    sock.close()

    logNetwork(db, ip, "TCP", port, state, status_id)
    return is_open
